// Dr. Thomas Chust/GenISys-GmbH
// memmap.cpp
#include <string>
#include <iostream>
#include <Windows.h>
#include "memmap.h"

namespace memmap {
    void MMapBuffer::Size2DWORDs(DWORD& high_p, DWORD& low_p, const size_t size_p)
    {
        if (sizeof(size_t) > 4) {
            high_p = DWORD((size_p >> 32) & 0xFFFFFFFF);
            low_p = DWORD(size_p & 0xFFFFFFFF);
        }
        else {
            high_p = 0;
            low_p = DWORD(size_p);
        }
    }

    size_t MMapBuffer::PageSize()
    {
        SYSTEM_INFO info;
        GetSystemInfo(&info);
        return size_t(info.dwAllocationGranularity);
    }

    void MMapBuffer::ThrowLastError()
    {
        DWORD error = GetLastError();
        std::string buf(8192, 0);
        const DWORD size = FormatMessageA(
            FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
            0, error,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            &buf[0], DWORD(buf.size()), 0
        );
        if (size > 0) {
            buf.resize(size);
            throw std::runtime_error(buf);
        }
        else {
            throw std::runtime_error("System error");
        }
    }

    MMapBuffer::MMapBuffer(const char* path_p, const size_t begin_p, const size_t size_p, const Options options_p) :
        m_File(INVALID_HANDLE_VALUE), m_Mapping(INVALID_HANDLE_VALUE),
        m_Data(0), m_Begin(begin_p), m_Size(0), m_Page(begin_p % PageSize())
    {
        const size_t pathlen = strlen(path_p);
        if (pathlen>0) {
            int result;

            // get length of converted string in characters
            result = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, (char *)path_p,
                (int)pathlen+1, NULL, 0);

            wchar_t wpath[1024];

            // convert string
            result = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, (char *)path_p,
                (int)pathlen+1, wpath, result);

            m_File = CreateFileW(
                wpath,
                GENERIC_READ | (options_p.m_ReadOnly ? 0 : GENERIC_WRITE),
                (options_p.m_ReadOnly ? FILE_SHARE_READ : 0),
                0,
                (options_p.m_Create ? OPEN_ALWAYS : OPEN_EXISTING),
                FILE_ATTRIBUTE_NORMAL,
                0
            );
            if (m_File == INVALID_HANDLE_VALUE) ThrowLastError();

            LARGE_INTEGER end;
            if (!GetFileSizeEx(m_File, &end)) ThrowLastError();

            if (size_p == size_t(-1)) {
                if (LONGLONG(m_Begin) < end.QuadPart) m_Size = size_t(end.QuadPart - LONGLONG(begin_p));
            }
            else if (LONGLONG(m_Begin) + LONGLONG(size_p) <= end.QuadPart) {
                m_Size = size_p;
            }
            else if (options_p.m_Create && size_p > 0) {
                end.QuadPart = LONGLONG(m_Begin) + LONGLONG(size_p);
                if (!SetFilePointerEx(m_File, end, 0, FILE_BEGIN)) ThrowLastError();
                if (!SetEndOfFile(m_File)) ThrowLastError();
                m_Size = size_p;
            }
        }

        if (m_File != INVALID_HANDLE_VALUE && m_Size > 0) {
            m_Begin -= m_Page;
            m_Size += m_Page;

            DWORD high, low;
            Size2DWORDs(high, low, m_Begin + m_Size);
            m_Mapping = CreateFileMappingW(
                m_File, 0, (options_p.m_ReadOnly ? PAGE_READONLY : PAGE_READWRITE),
                high, low,
                0
            );
            if (m_Mapping == INVALID_HANDLE_VALUE) ThrowLastError();
            Size2DWORDs(high, low, m_Begin);
            m_Data = MapViewOfFile(
                m_Mapping, (options_p.m_ReadOnly ? FILE_MAP_READ : FILE_MAP_ALL_ACCESS),
                high, low,
                SIZE_T(m_Size)
            );
            if (!m_Data) ThrowLastError();
        }
    }

    MMapBuffer::~MMapBuffer() /*override*/
    {
        if (m_Data) {
            if (!UnmapViewOfFile(m_Data)) ThrowLastError();
        }
        if (m_Mapping != INVALID_HANDLE_VALUE) {
            if (!CloseHandle(m_Mapping)) ThrowLastError();
        }
        if (m_File != INVALID_HANDLE_VALUE) {
            if (!CloseHandle(m_File)) ThrowLastError();
        }
    }

    void* MMapBuffer::data() const /*override*/
    {
        return static_cast< char* >(m_Data) + m_Page;
    }

    size_t MMapBuffer::size() const /*override*/
    {
        return m_Size - m_Page;
    }
};
