// vheap.cpp
//
#include "vheap.h"

struct spinlock
{
    spinlock(std::atomic<bool>& rmutex_p) :
        rmutex(rmutex_p)
    {
        while(std::atomic_exchange_explicit(&rmutex, true, std::memory_order_acquire))
            ; // spin until acquired
    }

    ~spinlock()
    {
        std::atomic_store_explicit(&rmutex, false, std::memory_order_release);
    }

    std::atomic<bool>& rmutex;
};

vheap::vheap() :
    idxBuf(0),
    segstart(0),
    segend(0),
    memcounter(0),
    freebase(0),
    spinmutex(false)
{
}

//allocate next memory block
void* vheap::allocateNextMemBlock(size_t size)
{
    spinlock lock(spinmutex);

    _SBLOCK *prevBlock = 0;
    _SBLOCK *freeBlock = freebase;

    while (freeBlock != 0)
    {
        // Check if the block size is sufficient, but less than twice the requested amount
        if ((freeBlock->size >= size)&&((freeBlock->size>>2)<size))
        {
            if (prevBlock != 0)
                prevBlock->next = freeBlock->next;
            else
                freebase = freeBlock->next;

            freeBlock->next = 0;

            return(void*)(((char*)freeBlock) + BLOCK_SIZE);
        }

        prevBlock = freeBlock;
        freeBlock = freeBlock->next;
    }

    _SBLOCK *newblock = (_SBLOCK*)sbrk(BLOCK_SIZE + size);
    {
        newblock->next = NULL;
        newblock->size = size;
    }

    return (void*)(((char*)newblock) + BLOCK_SIZE);
}

//free memory block
void vheap::freeMemBlock(_SBLOCK *head)
{
    spinlock lock(spinmutex);

    if (head != NULL)
    {
        head->next = freebase;
        freebase = head;
    }
}

//increment heap counter
void* vheap::sbrk(size_t size_p)
{
    char* memcounter_ = memcounter;
    memcounter += size_p;
    if (memcounter >= segend)
    {
        // move remaining gap to the freelist
        if (segend - memcounter_ > BLOCK_SIZE)
        {
            _SBLOCK *sblock = (_SBLOCK *)memcounter_;
            sblock->size = segend - memcounter_ - BLOCK_SIZE;
            sblock->next = freebase;
            freebase = sblock;
        }

        try
        {
            char filename[2048];
            sprintf(filename, VFILE_NAME, idxBuf);

            m_Buffers[idxBuf] = new((void*)m_SomeSpace) memmap::MMapBuffer(filename, 0, VFILE_SIZE, memmap::Options(false));

            segstart = (char*)m_Buffers[idxBuf]->data();
            segend = segstart + m_Buffers[idxBuf]->size();

            ++idxBuf;
            memcounter_ = segstart;
            memcounter = segstart + size_p;
        }
        catch (...)
        {
            exit(1);
        }
    }

    return (void*)memcounter_;
}

void vheap::vunlock()
{
    for (size_t idx(0); idx < idxBuf; ++idx)
    {
        VirtualUnlock(m_Buffers[idx]->data(), m_Buffers[idx]->size());
    }
}
